from threading import Thread
import pandas as pd
import time
import os

from .util import load, dump, is_valid_filename
from .core import MiraiSeeker, Ensembler

class HyperSearchSpace:
    """
    This class represents the search space of hyperparameters for a base model.

    :type model_class: type
    :param model_class: Any class that represents a statistical model. It must
        implement the methods ``fit`` as well as ``predict`` for regression or
        ``predict_proba`` for classification problems.

    :type id: str
    :param id: The id that will be associated with the models generated within
        this search space.

    :type parameters_values: dict, optional, default={ }
    :param parameters_values: A dictionary containing lists of values to be
        tested as parameters when instantiating objects of ``model_class``.

    :type parameters_rules: function, optional, default=lambda x: None
    :param parameters_rules: A function that constrains certain parameters because
        of the values assumed by others. It must receive a dictionary as input and
        doesn't need to return anything. Not used if ``parameters_values`` has no
        keys.

        .. warning::
             Make sure that the parameters accessed in ``parameters_rules`` exist
             in the set of parameters defined on ``parameters_values``, otherwise
             the engine will attempt to access an invalid key.

    :raises: ``NotImplementedError``, ``TypeError``, ``ValueError``

    :Example:

    ::

        from sklearn.linear_model import LogisticRegression
        from miraiml import HyperSearchSpace

        def logistic_regression_parameters_rules(parameters):
            if parameters['solver'] in ['newton-cg', 'sag', 'lbfgs']:
                parameters['penalty'] = 'l2'

        hyper_search_space = HyperSearchSpace(
            model_class = LogisticRegression,
            id = 'Logistic Regression',
            parameters_values = {
                'penalty': ['l1', 'l2'],
                'C': np.arange(0.1, 2, 0.1),
                'max_iter': np.arange(50, 300),
                'solver': ['newton-cg', 'lbfgs', 'liblinear', 'sag', 'saga'],
                'random_state': [0]
            },
            parameters_rules = logistic_regression_parameters_rules
        )

    .. warning::
        **Do not** allow ``random_state`` assume multiple values. If ``model_class``
        has a ``random_state`` parameter, force the engine to always choose the
        same value by providing a list with a single element.

        Allowing ``random_state`` to assume multiple values will confuse the engine
        because the scores will be unstable even with the same choice of
        hyperparameters and features.
    """
    def __init__(self, model_class, id, parameters_values={},
            parameters_rules=lambda x: None):
        self.__validate__(model_class, id, parameters_values, parameters_rules)
        self.model_class = model_class
        self.id = id
        self.parameters_values = parameters_values
        self.parameters_rules = parameters_rules

    def __validate__(self, model_class, id, parameters_values, parameters_rules):
        dir_model_class = dir(model_class)
        if 'fit' not in dir_model_class:
            raise NotImplementedError('model_class must implement fit')
        if type(id) != str:
            raise TypeError('id must be a string')
        if not is_valid_filename(id):
            raise ValueError('Invalid id: {}'.format(id))
        if type(parameters_values) != dict:
            raise TypeError('parameters_values must be a dictionary')
        if type(parameters_rules) != type(lambda: None):
            raise TypeError('parameters_rules must be a function')

class Config:
    """
    This class defines the general behavior of the engine.

    :type local_dir: str
    :param local_dir: The path for the engine to save its internal files. If the
        directory doesn't exist, it will be created automatically.

    :type problem_type: str
    :param problem_type: ``'classification'`` or ``'regression'``. The problem
        type. Multi-class classification problems are not supported.

    :type hyper_search_spaces: list
    :param hyper_search_spaces: The list of :class:`miraiml.HyperSearchSpace`
        objects to optimize. If ``hyper_search_spaces`` has length 1, the engine
        will not run ensemble cycles.

    :type score_function: function
    :param score_function: A function that receives the "truth" and the predictions
        (in this order) and returns the score. Bigger scores must mean better models.

    :type n_folds: int, optional, default=5
    :param n_folds: The number of folds for the fitting/predicting process.

    :type stratified: bool, optional, default=True
    :param stratified: Whether to stratify folds on target or not. Only used if
        ``problem_type == 'classification'``.

    :type ensemble_id: str, optional, default=None
    :param ensemble_id: The id for the ensemble. If none is given, the engine will
        not ensemble base models.

    :raises: ``NotImplementedError``, ``TypeError``, ``ValueError``

    :Example:

    ::

        from sklearn.metrics import roc_auc_score
        from miraiml import Config

        config = Config(
            local_dir = 'miraiml_local',
            problem_type = 'classification',
            hyper_search_spaces = hyper_search_spaces,
            score_function = roc_auc_score,
            n_folds = 5,
            stratified = True,
            ensemble_id = 'Ensemble'
        )
    """
    def __init__(self, local_dir, problem_type, hyper_search_spaces, score_function,
            n_folds=5, stratified=True, ensemble_id=None):
        self.__validate__(local_dir, problem_type, hyper_search_spaces, score_function,
                          n_folds, stratified, ensemble_id)
        self.local_dir = local_dir
        if self.local_dir[-1] != '/':
            self.local_dir += '/'
        self.problem_type = problem_type
        self.hyper_search_spaces = hyper_search_spaces
        self.score_function = score_function
        self.n_folds = n_folds
        self.stratified = stratified
        self.ensemble_id = ensemble_id

    def __validate__(self, local_dir, problem_type, hyper_search_spaces, score_function,
            n_folds, stratified, ensemble_id):
        if type(local_dir) != str:
            raise TypeError('local_dir must be a string')

        for dir_name in local_dir.split('/'):
            if not is_valid_filename(dir_name):
                raise ValueError('Invalid directory name: {}'.format(dir_name))

        if type(problem_type) != str:
            raise TypeError('problem_type must be a string')
        if problem_type not in ('classification', 'regression'):
            raise ValueError('Invalid problem type')

        if type(hyper_search_spaces) != list:
            raise TypeError('hyper_search_spaces must be a list')
        if len(hyper_search_spaces) == 0:
            raise ValueError('No search spaces')

        ids = []
        for hyper_search_space in hyper_search_spaces:
            if type(hyper_search_space) != HyperSearchSpace:
                raise TypeError('All hyper search spaces must be objects of '+\
                    'miraiml.HyperSearchSpace')
            id = hyper_search_space.id
            if id in ids:
                raise ValueError('Duplicated search space id: {}'.format(id))
            ids.append(id)
            dir_model_class = dir(hyper_search_space.model_class)
            if problem_type == 'classification' and 'predict_proba' not in dir_model_class:
                raise NotImplementedError('Model class of id {} '.format(id)+\
                    'must implement predict_proba for classification problems')
            if problem_type == 'regression' and 'predict' not in dir_model_class:
                raise NotImplementedError('Model class of id {} '.format(id)+\
                    'must implement predict for regression problems')

        if type(score_function) != type(lambda: None):
            raise TypeError('score_function must be a function')

        if type(n_folds) != int:
            raise TypeError('n_folds must be an integer')
        if n_folds < 2:
            raise ValueError('n_folds greater than 1')

        if type(stratified) != bool:
            raise TypeError('stratified must be a boolean')

        if type(ensemble_id) != type(None) and type(ensemble_id) != str:
            raise TypeError('ensemble_id must be a None or a string')
        if type(ensemble_id) == str and not is_valid_filename(ensemble_id):
            raise ValueError('invalid ensemble_id')
        if ensemble_id in ids:
            raise ValueError('ensemble_id cannot have the same id of a hyper '+\
                'search space')


class Engine:
    """
    This class offers the controls for the engine.

    :type config: miraiml.Config
    :param config: The configurations for the behavior of the engine.

    :type on_improvement: function, optional, default=None
    :param on_improvement: A function that will be executed everytime the engine
        finds an improvement for some id. It must receive a ``status`` parameter,
        which is the return of the method :func:`request_status`.

    :raises: ``TypeError``

    :Example:

    ::

        from miraiml import Engine

        def on_improvement(status):
            print('Scores:', status['scores'])

        engine = Engine(config, on_improvement=on_improvement)
    """
    def __init__(self, config, on_improvement=None):
        self.__validate__(config, on_improvement)
        self.config = config
        self.on_improvement = on_improvement
        self.__is_running__ = False
        self.must_interrupt = False
        self.mirai_seeker = None
        self.models_dir = config.local_dir + 'models/'
        self.train_data = None
        self.ensembler = None

    def __validate__(self, config, on_improvement):
        if type(config) != Config:
            raise TypeError('miraiml.Engine\'s constructor requires an object'+\
                ' of miraiml.Config')
        if type(on_improvement) != type(lambda: None) and\
            type(on_improvement) != type(None):
            raise TypeError('on_improvement must be None or a function')

    def is_running(self):
        """
        Tells whether the engine is running or not.

        :rtype: bool
        :returns: ``True`` if the engine is running and ``False`` otherwise.
        """
        return self.__is_running__

    def interrupt(self):
        """
        Makes the engine stop on the first opportunity.

        .. note::
            This method is **not** asynchronous. It will wait for the engine to
            stop.
        """
        self.must_interrupt = True
        if self.ensembler is not None:
            self.ensembler.interrupt()
        while self.__is_running__:
            time.sleep(.1)
        self.must_interrupt = False

    def load_data(self, train_data, target_column, test_data=None, restart=False):
        """
        Interrupts the engine and loads a new pair of train/test datasets.

        :type train_data: pandas.DataFrame
        :param train_data: The training data.

        :type target_column: str
        :param target_column: The name of the target column.

        :type test_data: pandas.DataFrame, optional, default=None
        :param test_data: The testing data. Use the default value if you don't
            need to make predictions for data with unknown labels.

        :type restart: bool, optional, default=False
        :param restart: Whether to restart the engine after updating data or not.
        """
        if type(train_data) != pd.DataFrame:
            raise TypeError('Training data must be an object of pandas.DataFrame')

        if type(test_data) != type(None) and type(test_data) != pd.DataFrame:
            raise TypeError('Testing data must be None or an object of pandas.DataFrame')

        self.interrupt()
        self.train_data = train_data
        self.train_target = self.train_data.pop(target_column)
        self.all_features = list(train_data.columns)
        self.test_data = test_data
        if self.mirai_seeker is not None:
            self.mirai_seeker.reset()
        if restart:
            self.restart()

    def shuffle_train_data(self, restart=False):
        """
        Interrupts the engine and shuffles the training data.

        :type restart: bool, optional, default=False
        :param restart: Whether to restart the engine after shuffling data or not.

        :raises: ``RuntimeError``

        .. note::
            It's a good practice to shuffle the training data periodically to avoid
            overfitting on a certain folding pattern.
        """
        if self.train_data is None:
            raise RuntimeError('No data to shuffle')

        self.interrupt()

        seed = int(time.time())
        self.train_data = self.train_data.sample(frac=1, random_state=seed)
        self.train_target = self.train_target.sample(frac=1, random_state=seed)

        if restart:
            self.restart()

    def reconfigure(self, config, restart=False):
        """
        Interrupts the engine and loads a new configuration.

        :type config: miraiml.Config
        :param config: The configurations for the behavior of the engine.

        :type restart: bool, optional, default=False
        :param restart: Whether to restart the engine after reconfiguring it or
            not.
        """
        self.interrupt()
        self.config = config
        if self.mirai_seeker is not None:
            self.mirai_seeker.reset()
        if restart:
            self.restart()

    def restart(self):
        """
        Interrupts the engine and starts again from last checkpoint (if any).

        :raises: ``RuntimeError``, ``KeyError``
        """
        if self.train_data is None:
            raise RuntimeError('No data to train')
        self.interrupt()

        def starter():
            try:
                self.__main_loop__()
            except:
                self.__is_running__ = False
                raise

        Thread(target=lambda: starter()).start()

    def __improvement_trigger__(self):
        """
        Called when an improvement happens.
        """
        if self.on_improvement is not None:
            self.on_improvement(self.request_status())

    def __check_best__(self, score, id):
        if self.best_score is None or score > self.best_score:
            self.best_score = score
            self.best_id = id

    def __main_loop__(self):
        """
        Main optimization loop.
        """
        self.__is_running__ = True
        if not os.path.exists(self.models_dir):
            os.makedirs(self.models_dir)
        self.base_models = {}
        self.train_predictions_df = pd.DataFrame()
        self.test_predictions_df = pd.DataFrame()
        self.scores = {}
        self.best_score = None
        self.best_id = None

        self.mirai_seeker = MiraiSeeker(
            self.config.hyper_search_spaces,
            self.all_features,
            self.config
        )

        start = time.time()
        for hyper_search_space in self.config.hyper_search_spaces:
            if self.must_interrupt:
                break
            id = hyper_search_space.id
            base_model_path = self.models_dir + id
            if os.path.exists(base_model_path):
                base_model = load(base_model_path)
            else:
                base_model = self.mirai_seeker.seek(hyper_search_space.id)
                dump(base_model, base_model_path)
            self.base_models[id] = base_model

            train_predictions, test_predictions, score = base_model.predict(
                self.train_data, self.train_target, self.test_data, self.config)
            self.train_predictions_df[id] = train_predictions
            self.test_predictions_df[id] = test_predictions
            self.scores[id] = score

            self.__check_best__(self.scores[id], id)

        total_cycles_duration = time.time() - start
        n_cycles = 1

        will_ensemble = len(self.base_models) > 1 and\
            self.config.ensemble_id is not None

        if will_ensemble:
            self.ensembler = Ensembler(
                list(self.base_models),
                self.train_target,
                self.train_predictions_df,
                self.test_predictions_df,
                self.scores,
                self.config
            )

            ensemble_id = self.config.ensemble_id

            if self.ensembler.optimize(total_cycles_duration):
                self.__check_best__(self.scores[ensemble_id], ensemble_id)

        self.__improvement_trigger__()

        while not self.must_interrupt:

            start = time.time()
            for hyper_search_space in self.config.hyper_search_spaces:
                if self.must_interrupt:
                    break
                id = hyper_search_space.id

                base_model = self.mirai_seeker.seek(id)

                train_predictions, test_predictions, score = base_model.predict(
                    self.train_data, self.train_target,
                    self.test_data, self.config)

                self.mirai_seeker.register_base_model(id, base_model, score)

                if score > self.scores[id] or (score == self.scores[id] and\
                        len(base_model.features) < len(self.base_models[id].features)):
                    self.scores[id] = score
                    self.train_predictions_df[id] = train_predictions
                    self.test_predictions_df[id] = test_predictions
                    self.__check_best__(score, id)

                    if will_ensemble:
                        self.ensembler.update()
                        self.__check_best__(self.scores[ensemble_id], ensemble_id)

                    self.__improvement_trigger__()

                    dump(base_model, self.models_dir + id)

            total_cycles_duration += time.time() - start
            n_cycles += 1

            if will_ensemble:
                if self.ensembler.optimize(total_cycles_duration/n_cycles):
                    self.__check_best__(self.scores[ensemble_id], ensemble_id)

                    self.__improvement_trigger__()

        self.__is_running__ = False

    def request_status(self):
        """
        Queries the current status of the engine.

        :rtype: dict or None
        :returns: The current status of the engine in the form of a dictionary.
            If no score has been computed yet, returns ``None``. The available
            keys and their respective values on the status dictionary are:

            * ``'best_id'``: The current best id

            * ``'scores'``: A dictionary containing the score of each id

            * ``'predictions'``: A ``pandas.DataFrame`` object containing the\
                predictions from each id for the testing dataset. If no testing\
                dataset was provided, the value associated with this key is\
                ``None``

            * ``'ensemble_weights'``: A dictionary containing the ensemble weights\
                for each base model. If no ensembling cycle has been executed,\
                the value associated with this key is ``None``

            * ``'base_models'``: A dictionary containing the current description\
                of each base model, which can be accessed by their ids

            The dictionary associated with the ``'base_models'`` key contains the
            following keys and respective values:

            * ``'model_class'``: The name of the base model's class

            * ``'parameters'``: The dictionary of hyperparameters

            * ``'features'``: The list of features
        """
        if self.best_id is None:
            return None

        predictions = None
        if self.test_data is not None:
            predictions = self.test_predictions_df.copy()

        ensemble_weights = None
        if self.ensembler is not None:
            ensemble_weights = self.ensembler.weights.copy()

        base_models = {}
        for id in self.base_models:
            base_model = self.base_models[id]
            base_models[id] = dict(
                model_class = base_model.model_class.__name__,
                parameters = base_model.parameters.copy(),
                features = base_model.features.copy()
            )

        return dict(
            best_id = self.best_id,
            scores = self.scores.copy(),
            predictions = predictions,
            ensemble_weights = ensemble_weights,
            base_models = base_models
        )



"""
:mod:`miraiml.core` contains internal classes responsible for the optimization
process.

- :class:`miraiml.core.BaseModel` represents a solution
- :class:`miraiml.core.MiraiSeeker` implements the strategies to search for good
  solutions
- :class:`miraiml.core.Ensembler` searches for smart ways of combining the current
  solutions o generate a better one
"""

import random as rnd
import pandas as pd
import numpy as np
import time
import os

from sklearn.model_selection import StratifiedKFold, KFold
from sklearn.linear_model import LinearRegression

from .util import load, dump, sample_random_len

class BaseModel:
    """
    Represents an element from the search space, defined by an instance of
    :class:`miraiml.HyperSearchSpace` and a set of features.

    Read more in the :ref:`User Guide <base_model>`.

    :type model_class: type
    :param model_class: A statistical model class that must implement the methods
        ``fit`` and ``predict`` for regression or ``predict_proba`` classification
        problems.

    :type parameters: dict
    :param parameters: The parameters that will be used to instantiate objects of
        ``model_class``.

    :type features: list
    :param features: The list of features that will be used to train the statistical
        model.
    """
    def __init__(self, model_class, parameters, features):
        self.model_class = model_class
        self.parameters = parameters
        self.features = features

    def predict(self, X_train, y_train, X_test, config):
        """
        Performs the predictions for the training and testing datasets and also
        computes the score of the model.

        :type X_train: pandas.DataFrame
        :param X_train: The dataframe that contains the training inputs for the
            model.

        :type y_train: pandas.Series or numpy.ndarray
        :param y_train: The training targets for the model.

        :type X_test: pandas.DataFrame
        :param X_test: The dataframe that contains the testing inputs for the model.

        :type config: miraiml.Config
        :param config: The configuration of the engine.

        :rtype: tuple
        :returns: ``(train_predictions, test_predictions, score)``

            * ``train_predictions``: The predictions for the training dataset
            * ``test_predictions``: The predictions for the testing dataset
            * ``score``: The score of the model on the training dataset

        :raises: ``RuntimeError``
        """
        X_train = X_train[self.features]
        train_predictions = np.zeros(X_train.shape[0])

        test_predictions = None
        if X_test is not None:
            X_test = X_test[self.features]
            test_predictions = np.zeros(X_test.shape[0])

        if config.problem_type == 'classification' and config.stratified:
            fold = StratifiedKFold(n_splits=config.n_folds, shuffle=False)
        elif config.problem_type == 'regression' or not config.stratified:
            fold = KFold(n_splits=config.n_folds, shuffle=False)
        for big_part, small_part in fold.split(X_train, y_train):

            X_train_big, y_train_big = X_train.iloc[big_part], y_train.iloc[big_part]
            X_train_small = X_train.iloc[small_part]

            model = self.model_class(**self.parameters)
            class_name = self.model_class.__name__

            try:
                model.fit(X_train_big, y_train_big)
            except:
                raise RuntimeError('Error when fitting with model class {}'.\
                    format(class_name))
            try:
                if config.problem_type == 'classification':
                    train_predictions[small_part] = model.predict_proba(X_train_small)[:,1]
                    if X_test is not None:
                        test_predictions += model.predict_proba(X_test)[:,1]
                elif config.problem_type == 'regression':
                    train_predictions[small_part] = model.predict(X_train_small)
                    if X_test is not None:
                        test_predictions += model.predict(X_test)
            except:
                raise RuntimeError('Error when predicting with model class {}'.\
                    format(class_name))

        if X_test is not None:
            test_predictions /= config.n_folds
        return (train_predictions, test_predictions,
                config.score_function(y_train, train_predictions))

class MiraiSeeker:
    """
    This class implements a smarter way of searching good parameters and sets of
    features.

    Read more in the :ref:`User Guide <mirai_seeker>`.

    :param base_models_ids: The list of base models' ids to keep track of.
    :type base_models_ids: list

    :param all_features: A list containing all available features.
    :type all_features: list

    :param config: The configuration of the engine.
    :type config: miraiml.Config
    """
    def __init__(self, hyper_search_spaces, all_features, config):
        self.all_features = all_features
        self.config = config

        histories_path = config.local_dir + 'histories/'

        if not os.path.exists(histories_path):
            os.makedirs(histories_path)

        self.hyper_search_spaces_dict = {}
        self.histories = {}
        self.histories_paths = {}
        for hyper_search_space in hyper_search_spaces:
            id = hyper_search_space.id
            self.hyper_search_spaces_dict[id] = hyper_search_space

            self.histories_paths[id] = histories_path + id
            if os.path.exists(self.histories_paths[id]):
                self.histories[id] = load(self.histories_paths[id])
            else:
                self.histories[id] = pd.DataFrame()
                dump(self.histories[id], self.histories_paths[id])

    def reset(self):
        """
        Deletes all base models registries.
        """
        for id in self.hyper_search_spaces_dict:
            self.histories[id] = pd.DataFrame()
            dump(self.histories[id], self.histories_paths[id])

    def parameters_features_to_dataframe(self, parameters, features, score):
        """
        Creates an entry for a history.

        :type parameters: dict
        :param parameters: The set of parameters to transform.

        :type parameters: list
        :param parameters: The set of features to transform.

        :type score: float
        :param score: The score to transform.
        """
        entry = {'score':score}
        for parameter in parameters:
            entry[parameter+'(parameter)'] = parameters[parameter]
        for feature in self.all_features:
            entry[feature+'(feature)'] = 1 if feature in features else 0
        return pd.DataFrame([entry])

    def register_base_model(self, id, base_model, score):
        """
        Registers the performance of a base model and its characteristics.

        :type id: str
        :param id: The id associated with the base model.

        :type base_model: miraiml.core.BaseModel
        :param base_model: The base model being registered.

        :type score: float
        :param score: The score of ``base_model``.
        """
        new_entry = self.parameters_features_to_dataframe(
            base_model.parameters,
            base_model.features, score)

        self.histories[id] = pd.concat([self.histories[id], new_entry])
        self.histories[id].drop_duplicates(inplace=True)
        dump(self.histories[id], self.histories_paths[id])

    def is_ready(self, id):
        """
        Tells whether the history of an id is large enough for more advanced
        strategies or not.

        :type id: str
        :param id: The id to be inspected.

        :rtype: bool
        :returns: Whether ``id`` can be used to generate parameters and features
            lists or not.
        """
        return self.histories[id].shape[0] > 1

    def seek(self, id):
        """
        Manages the search strategy for better solutions.

        With a probability of 0.5, the random strategy will be chosen. If it's
        not, the other strategies will be chosen with equal probabilities.

        :type id: str
        :param id: The id for which a new base model is required.

        :rtype: miraiml.core.BaseModel
        :returns: The next base model for exploration.

        :raises: ``KeyError``
        """
        if rnd.uniform(0, 1) > 0.5 or not self.is_ready(id):
            parameters, features = self.random_search(id)
        else:
            if rnd.uniform(0, 1) > 0.5:
                parameters, features = self.naive_search(id)
            else:
                parameters, features = self.linear_regression_search(id)

        hyper_search_space = self.hyper_search_spaces_dict[id]
        if len(parameters) > 0:
            try:
                hyper_search_space.parameters_rules(parameters)
            except:
                raise KeyError('Error on parameters rules for the id {}'.format(id))

        model_class = hyper_search_space.model_class

        return BaseModel(model_class, parameters, features)

    def random_search(self, id):
        """
        Generates completely random sets of parameters and features.

        :type all_features: list
        :param all_features: The list of available features.

        :rtype: tuple
        :returns: ``(parameters, features)``
            Respectively, the dictionary of parameters and the list of features
            that can be used to generate a new base model.
        """
        hyper_search_space = self.hyper_search_spaces_dict[id]
        model_class = hyper_search_space.model_class
        parameters = {}
        for parameter in hyper_search_space.parameters_values:
            parameters[parameter] = rnd.choice(
                hyper_search_space.parameters_values[parameter])
        features = sample_random_len(self.all_features)
        return (parameters, features)

    def naive_search(self, id):
        """
        Characteristics that achieved higher scores have independently higher
        chances of being chosen again.

        :type id: str
        :param id: The id for which we want a new set of parameters and features.

        :rtype: tuple
        :returns: ``(parameters, features)``
            Respectively, the dictionary of parameters and the list of features
            that can be used to generate a new base model.
        """
        history = self.histories[id]
        parameters = {}
        features = []
        for column in history.columns:
            if column == 'score':
                continue
            dist = history[[column, 'score']].groupby(column).mean().reset_index()
            chosen_value = rnd.choices(
                dist[column].values,
                cum_weights=dist['score'].cumsum().values)[0]
            if column.endswith('(parameter)'):
                parameter = column.split('(')[0]
                parameters[parameter] = chosen_value
            elif column.endswith('(feature)'):
                feature = column.split('(')[0]
                if chosen_value:
                    features.append(feature)
        if len(features) == 0:
            features = sample_random_len(self.all_features)
        return (parameters, features)

    def dataframe_to_parameters_features(self, dataframe):
        """
        Transforms a history entry in a pair of parameters and features.

        :type dataframe: pandas.DataFrame
        :param dataframe: The history entry to be transformed,

        :rtype: tuple
        :returns: ``(parameters, features)``. The transformed sets of parameters
            and features, respectively.
        """
        parameters = {}
        features = []
        for column in dataframe.columns:
            if column == 'score':
                continue
            column_filtered = column.split('(')[0]
            value = dataframe[column].values[0]
            if column.endswith('(parameter)'):
                parameters[column_filtered] = value
            elif column.endswith('(feature)'):
                if value:
                    features.append(column_filtered)
        return (parameters, features)

    def linear_regression_search(self, id):
        """
        Uses the history to model the score with a linear regression. Guesses the
        scores of `n`/2 random sets of parameters and features, where `n` is the
        size of the history. The one with the highest score is chosen.

        :type id: str
        :param id: The id for which we want a new set of parameters and features.

        :rtype: tuple
        :returns: ``(parameters, features)``
            Respectively, the dictionary of parameters and the list of features
            that can be used to generate a new base model.
        """
        history = self.histories[id]
        n_guesses = history.shape[0]//2

        # Creating guesses:
        guesses_df = pd.DataFrame()
        for _ in range(n_guesses):
            guess_parameters, guess_features = self.random_search(id)
            guess_df = self.parameters_features_to_dataframe(
                guess_parameters, guess_features, np.nan)
            guesses_df = pd.concat([guesses_df, guess_df])

        # Concatenating data to perform one-hot-encoding:
        data = pd.concat([history, guesses_df])
        object_columns = [col for col in data.columns if data[col].dtype == object]
        data_ohe = pd.get_dummies(data, columns=object_columns, drop_first=True)

        # Separating train and test:
        train_mask = data_ohe['score'].notna()
        data_ohe_train = data_ohe[train_mask]
        data_ohe_test = data_ohe[~train_mask].drop(columns='score')
        y = data_ohe_train.pop('score')

        # Fitting and predicting scores:
        model = LinearRegression(normalize=True)
        model.fit(data_ohe_train, y)
        guesses_df['score'] = model.predict(data_ohe_test)

        # Choosing the best guess:
        best_guess = guesses_df.sort_values('score', ascending=False).head(1)

        return self.dataframe_to_parameters_features(best_guess)

class Ensembler:
    """
    Performs the ensemble of the base models.

    Read more in the :ref:`User Guide <ensemble>`.

    :type y_train: pandas.Series or numpy.ndarray
    :param y_train: The target column.

    :type base_models_ids: list
    :param base_models_ids: The list of base models' ids to keep track of.

    :type train_predictions_df: pandas.DataFrame
    :param train_predictions_df: The dataframe of predictions for the training
        dataset.

    :type test_predictions_df: pandas.DataFrame
    :param test_predictions_df: The dataframe of predictions for the testing
        dataset.

    :type scores: dict
    :param scores: The dictionary of scores.

    :type config: miraiml.Config
    :param config: The configuration of the engine.
    """
    def __init__(self, base_models_ids, y_train, train_predictions_df,
            test_predictions_df, scores, config):
        self.y_train = y_train
        self.base_models_ids = sorted(base_models_ids)
        self.train_predictions_df = train_predictions_df
        self.test_predictions_df = test_predictions_df
        self.scores = scores
        self.config = config
        self.id = config.ensemble_id
        self.weights_path = config.local_dir + 'models/' + self.id
        self.must_interrupt = False

        if os.path.exists(self.weights_path):
            self.weights = load(self.weights_path)
        else:
            self.weights = self.gen_weights()
            dump(self.weights, self.weights_path)

    def interrupt(self):
        """
        Sets an internal flag to interrupt the optimization process on the first
        opportunity.
        """
        self.must_interrupt = True

    def update(self):
        """
        Updates the ensemble with the newest predictions from the base models.
        """
        train_predictions, test_predictions, score = self.ensemble(self.weights)
        self.train_predictions_df[self.id] = train_predictions
        self.test_predictions_df[self.id] = test_predictions
        self.scores[self.id] = score

    def gen_weights(self):
        """
        Generates the ensemble weights according to the score of each base model.
        Higher scores have higher chances of generating higher weights.

        :rtype: dict
        :returns: A dictionary containing the weights for each base model id.
        """
        weights = {}
        if len(self.scores) > 0:
            min_score, max_score = np.inf, -np.inf
            for id in self.base_models_ids:
                score = self.scores[id]
                min_score = min(min_score, score)
                max_score = max(max_score, score)
            diff_score = max_score - min_score
            for id in self.base_models_ids:
                if self.scores[id] == max_score:
                    weights[id] = rnd.triangular(0, 1, 1)
                else:
                    normalized_score = (self.scores[id]-min_score)/diff_score
                    range = rnd.triangular(0, 1, normalized_score)
                    weights[id] = rnd.triangular(0, range, 0)
        else:
            for id in self.base_models_ids:
                weights[id] = 1
        return weights

    def ensemble(self, weights):
        """
        Performs the ensemble of the current predictions of each base model.

        :type weights: dict
        :param weights: A dictionary containing the weights related to the id of
            each base model.

        :rtype: tuple
        :returns: ``(train_predictions, test_predictions, score)``

            * ``train_predictions``: The ensemble predictions for the training dataset
            * ``test_predictions``: The ensemble predictions for the testing dataset
            * ``score``: The score of the ensemble on the training dataset
        """
        weights_list = [weights[id] for id in self.base_models_ids]
        train_predictions = np.average(
            self.train_predictions_df[self.base_models_ids],
            axis=1, weights=weights_list)
        test_predictions = None
        if self.test_predictions_df.shape[0] > 0:
            test_predictions = np.average(
                self.test_predictions_df[self.base_models_ids],
                axis=1, weights=weights_list)
        return (train_predictions, test_predictions,
                self.config.score_function(self.y_train, train_predictions))

    def optimize(self, max_duration):
        """
        Performs ensembling cycles for ``max_duration`` seconds.

        :type max_duration: float
        :param max_duration: The maximum duration allowed for the optimization
            process.

        :rtype: bool
        :returns: ``True`` if a better set of weights was found and ``False``
            otherwise.
        """
        optimized = False
        start = time.time()
        while time.time() - start < max_duration and not self.must_interrupt:
            weights = self.gen_weights()
            train_predictions, test_predictions, score = self.ensemble(weights)
            if self.id not in self.scores or score > self.scores[self.id]:
                self.scores[self.id] = score
                self.weights = weights
                self.train_predictions_df[self.id] = train_predictions
                self.test_predictions_df[self.id] = test_predictions
                dump(self.weights, self.weights_path)
                optimized = True
        return optimized



"""
:mod:`miraiml` provides the following classes:

- :class:`miraiml.HyperSearchSpace` represents the search space of hyperparameters
- :class:`miraiml.Config` defines the general behavior of :class:`miraiml.Engine`
- :class:`miraiml.Engine` manages the optimization process

You can import them by doing

>>> from miraiml import HyperSearchSpace, Config, Engine
"""

__version__ = '2.0.2'
__all__ = ['HyperSearchSpace', 'Config', 'Engine']

from .main import HyperSearchSpace, Config, Engine



from distutils.core import setup
from miraiml import __version__

VERSION = __version__
with open('README.rst') as f:
    LONG_DESCRIPTION = f.read()

setup(
    name = "MiraiML",
    version = VERSION,
    packages = ["miraiml"],
    author = "Arthur Paulino",
    author_email = "arthurleonardo.ap@gmail.com",
    url = "https://github.com/arthurpaulino/miraiml",
    description = "An asynchronous engine for continuous & autonomous machine\
        learning, built for real-time usage",
    long_description = LONG_DESCRIPTION
)



"""
:mod:`miraiml.util` provides utility functions that are used by higher level
modules.
"""

from threading import Thread
import random as rnd
import pickle
import string
import math

def load(path):
    """
    A clean `pickle.load` wrapper for binary files.

    :type path: string
    :param path: The path of the binary file to be loaded.

    :rtype: object
    :returns: The loaded object.
    """
    return pickle.load(open(path, 'rb'))

def dump(obj, path):
    """
    Optimizes the process of writing objects on disc by triggering a thread.

    :type obj: object
    :param obj: The object to be dumped to the binary file.

    :type path: string
    :param path: The path of the binary file to be written.
    """
    Thread(target=lambda: pickle.dump(obj, open(path, 'wb'))).start()

def sample_random_len(lst):
    """
    Returns a sample of random size from the list ``lst``. The minimum length of
    the returned list is 1.

    :type lst: list
    :param lst: A list containing the elements to be sampled.

    :rtype: sampled_lst: list
    :returns: The randomly sampled elements from ``lst``.
    """
    if len(lst) == 0:
        return []
    return rnd.sample(lst, max(1, math.ceil(rnd.random()*len(lst))))

__valid_chars__ = frozenset("-_.() %s%s" % (string.ascii_letters, string.digits))

def is_valid_filename(filename):
    """
    Tells whether a string can be used as a safe file name or not.

    :type filename: str
    :param filename: The file name.

    :rtype: bool
    :returns: Whether ``filename`` is a valid file name or not.
    """
    if len(filename) == 0:
        return False
    for char in filename:
        if char not in __valid_chars__:
            return False
    return True
